/* findbtn.c */

#include <Pilot.h>

#include <SysEvtMgr.h>		/* for EvtFlushNextPenStroke() */
#include <SysEvtPrv.h>		/* for EvtGetPenBtnList() */

#include "findbtn.h"


/*
 * macros
 */

#ifdef __GNUC__
register void *reg_a7 asm("%sp");
#define __getA7() (long) reg_a7
#define __setA7(a7) reg_a7 = (void *) (a7)
#else /* __MWERKS__ */
long __getA7(void) = { 0x200F };
void __setA7(long a7:__d0) = { 0x2E40 };
#endif

#define CENTER_X(r) ((r).topLeft.x + (r).extent.x / 2 + 2)
#define CENTER_Y(r) ((r).topLeft.y + (r).extent.y / 2)


/*
 * constants
 */

#define FTR_ORIG 1000

#define PANEL_DB_TYPE 'panl'
#define PREFS_CREATOR 'pref'

#define DA_RSRC_TYPE 'code'
#define DA_RSRC_ID 1000

#define LOC_DB_CREATOR 'JLoc'

#define MIN_STACK_SIZE 256


/*
 * functions
 */

static int find_part(PointType *start, PointType *end)
{
    PenBtnInfoPtr buttons;
    UInt num_buttons, i;
    RectanglePtr r = NULL;
    struct findbtn_prefs prefs;

    buttons = EvtGetPenBtnList(&num_buttons);
    for (i = 0; i < num_buttons; ++i) {
	if (buttons[i].asciiCode == findChr) {
	    r = &buttons[i].boundsR;
	    break;
	}
    }

    if (r != NULL
     && RctPtInRectangle(start->x, start->y, r)
     && RctPtInRectangle(end->x, end->y, r)) {

	get_prefs(&prefs);

	if (start->x < CENTER_X(*r)) {
	    if (start->y < CENTER_Y(*r)) {
		return prefs.button1;
	    } else {
		return prefs.button2;
	    }
	} else {
	    if (start->y < CENTER_Y(*r)) {
		return prefs.button3;
	    } else {
		return prefs.button4;
	    }
	}
    }

    return MODE_NONE;
}

static Boolean IsDeviceLocked()
{
    DWord rom_version;
    SystemPreferencesTypeV10 sys_prefs;

    if (FtrGet(sysFtrCreator, sysFtrNumROMVersion, &rom_version) != noErr) {
	return false;
    } else if (sysGetROMVerMajor(rom_version) < 2) {
	PrefGetPreferences((SystemPreferencesPtr) &sys_prefs);
	return sys_prefs.deviceLocked;
    } else {
	return PrefGetPreference(prefDeviceLocked);
    }
}

static ULong GetCurAppCreator()
{
    UInt card_no;
    LocalID db_id;
    ULong type, creator;

    if (SysCurAppDatabase(&card_no, &db_id) == noErr
     && DmDatabaseInfo(card_no, db_id, NULL, NULL, NULL, NULL,
	    NULL, NULL, NULL, NULL, NULL, &type, &creator) == noErr) {
	if (type == PANEL_DB_TYPE) {
	    return PREFS_CREATOR;
	} else {
	    return creator;
	}
    }

    return 0;
}


/*
 * launching
 */

static void launch_app(int part)
{
    struct findbtn_prefs prefs;
    ULong creator;
    DmSearchStateType state;
    UInt card_no;
    LocalID db_id;

    if (IsDeviceLocked() == false) {
	get_prefs(&prefs);
	creator = (part == MODE_APP1) ? prefs.app1 : prefs.app2;
	if (creator != 0
	 && creator != GetCurAppCreator()
	 && DmGetNextDatabaseByTypeCreator(true, &state,
		APP_DB_TYPE, creator, true, &card_no, &db_id) == noErr) {
	    SysUIAppSwitch(card_no, db_id, sysAppLaunchCmdNormalLaunch, 0);
	}
    }
}

static void call_da(void (*da_entry)(), ULong stack_size, ULong loc_creator)
{
    FormPtr saved_form;
    Word saved_form_id;
    WinHandle saved_active_win;
    WinHandle saved_draw_win;
    FontID saved_font;
    DmOpenRef loc_db = NULL;
    VoidPtr new_stack;
    long saved_stack;

    /*
     * save active form, etc.
     */

    saved_form = FrmGetActiveForm();
    saved_form_id = FrmGetActiveFormID();
    saved_active_win = WinGetActiveWindow();
    if ((void *) saved_form != (void *) saved_active_win) {
	return;
    }
    saved_draw_win = WinGetDrawWindow();
    saved_font = FntSetFont(stdFont);

    /*
     * open localizer
     */

    if (loc_creator != 0) {
	loc_db = DmOpenDatabaseByTypeCreator(
		loc_creator, LOC_DB_CREATOR, dmModeReadOnly);
    }

    /*
     * prepare stack
     */

    new_stack = NULL;
    stack_size &= ~1L;
    if (stack_size >= MIN_STACK_SIZE) {
	new_stack = MemPtrNew(stack_size);
    }
    saved_stack = __getA7();	/* to shut up gcc -W ... */
    if (new_stack != NULL) {
	__setA7((long) new_stack + stack_size);
    }

    /*
     * call the desk accessory
     */

    InsPtEnable(false);
    (*da_entry)();

    /*
     * clean up
     */

    if (new_stack != NULL) {
	__setA7(saved_stack);
	MemPtrFree(new_stack);
    }

    if (loc_db != NULL) {
	DmCloseDatabase(loc_db);
    }

    /*
     * restore active form, etc.
     */

    if (saved_form_id != 0) {
	FrmSetActiveForm(saved_form);
    }
    WinSetActiveWindow(saved_active_win);
    WinSetDrawWindow(saved_draw_win);
    FntSetFont(saved_font);
}

static void launch_da(int part)
{
    struct findbtn_prefs prefs;
    ULong creator;
    DmOpenRef db;
    UInt open_count;
    Boolean res_db;
    VoidHand h;
    void (*da_entry)();

    if (IsDeviceLocked() == false) {
	get_prefs(&prefs);
	creator = (part == MODE_DA1) ? prefs.da1 : prefs.da2;
	if (creator != 0
	 && (db = DmOpenDatabaseByTypeCreator(
		DA_DB_TYPE, creator, dmModeReadOnly)) != NULL) {
	    if (DmOpenDatabaseInfo(
		    db, NULL, &open_count, NULL, NULL, &res_db) == noErr
	     && open_count == 1
	     && res_db != false) {
		h = DmGet1Resource(DA_RSRC_TYPE, DA_RSRC_ID);
		if (h != NULL) {
		    da_entry = MemHandleLock(h);
		    if (da_entry != NULL) {
			call_da(da_entry, prefs.stack_size,
				prefs.localizer ? creator : 0);
			MemHandleUnlock(h);
		    }
		    DmReleaseResource(h);
		}
	    }
	    DmCloseDatabase(db);
	}
    }
}


/*
 * main routines
 */

static void post_event(int part)
{
    Word chr = 0;
    EventType event;

    switch (part) {
    case MODE_GRAFFITI:		chr = graffitiReferenceChr;	break;
    case MODE_KEYBOARD:		chr = keyboardChr;		break;
    case MODE_KEYBOARD_ABC:	chr = keyboardAlphaChr;		break;
    case MODE_KEYBOARD_123:	chr = keyboardNumericChr;	break;
    }

    if (chr != 0) {
	MemSet(&event, sizeof event, 0);
	event.eType = keyDownEvent;
	event.data.keyDown.chr = chr;
	event.data.keyDown.keyCode = 0;
	event.data.keyDown.modifiers = commandKeyMask;
	EvtAddEventToQueue(&event);
    }
}

Err MyHandler(PointType *start, PointType *end)
{
    int part;
    Err (*orig_trap)(PointType *start, PointType *end);

    part = find_part(start, end);
    if (part != MODE_NONE) {
	EvtFlushNextPenStroke();
	if (part == MODE_APP1 || part == MODE_APP2) {
	    launch_app(part);
	} else if (part == MODE_DA1 || part == MODE_DA2) {
	    launch_da(part);
	} else {
	    post_event(part);
	}
    } else if (FtrGet(CREATOR, FTR_ORIG, (DWord *) &orig_trap) == noErr) {
	return (*orig_trap)(start, end);
    }

    return noErr;
}
