/* findbtn_panel.c */

#include <Pilot.h>

#include "findbtn.h"
#include "findbtn_rcp.h"


/*
 * constants and macros
 */

#define FTR_GLOBALS 9000

#define HACKMASTER_FORM_ID 9000

#define BAD_STACK_SIZE 0xff

#define MAX_LIST_HEIGHT 14
#define LIST_HEIGHT(x) ((x) < MAX_LIST_HEIGHT ? (x) : MAX_LIST_HEIGHT)


/*
 * pseudo globals
 */

struct db_list {
    Word count;
    VoidHand h;
    SysDBListItemType *list;
    CharPtr *ptrs;
};

struct globals {
    struct db_list apps;
    struct db_list das;
};


/*
 * functions
 */

static VoidPtr GetObjectPtr(Word id)
{
    FormPtr form;

    form = FrmGetActiveForm();
    return FrmGetObjectPtr(form, FrmGetObjectIndex(form, id));
}

static void set_popup(Word popup_id, Word index)
{
    ControlPtr popup;
    ListPtr list;
    CharPtr text;

    popup = GetObjectPtr(popup_id);
    list = GetObjectPtr(popup_id + 1);		/* list_id = popup_id + 1 */

    if (index < LstGetNumberOfItems(list)) {
	text = LstGetSelectionText(list, index);
    } else {
	text = "???";
	index = noListSelection;
    }

    CtlSetLabel(popup, text);
    LstSetSelection(list, index);
}

static void set_pref(Word list_id, Byte *value_ptr)
{
    ListPtr list;
    Word index;

    list = GetObjectPtr(list_id);
    index = LstGetSelection(list);
    if (index != noListSelection) {
	*value_ptr = index;
    }
}

static void set_popups()
{
    struct findbtn_prefs prefs;
    Byte stack_size;

    get_prefs(&prefs);

    set_popup(PANEL_BUTTON1_POPUP_ID, prefs.button1);
    set_popup(PANEL_BUTTON2_POPUP_ID, prefs.button2);
    set_popup(PANEL_BUTTON3_POPUP_ID, prefs.button3);
    set_popup(PANEL_BUTTON4_POPUP_ID, prefs.button4);

    set_popup(PANEL_LOCALIZER_POPUP_ID, prefs.localizer);

    stack_size = prefs.stack_size / STACK_UNIT;
    if (stack_size * STACK_UNIT != prefs.stack_size) {
	stack_size = BAD_STACK_SIZE;
    }
    set_popup(PANEL_STACK_POPUP_ID, stack_size);
}

static void set_db_popup(Word popup_id, struct db_list *dbs, ULong creator)
{
    Word index = noListSelection;
    int i;

    if (dbs->count > 0 && dbs->list != NULL) {
	for (i = 0; i < dbs->count; ++i) {
	    if (dbs->list[i].creator == creator) {
		index = i;
		break;
	    }
	}
    }

    set_popup(popup_id, index);
}

static void set_db_popups()
{
    struct findbtn_prefs prefs;
    struct globals *g;

    get_prefs(&prefs);

    if (FtrGet(CREATOR, FTR_GLOBALS, (DWord *) &g) == noErr
     && g != NULL) {
	set_db_popup(PANEL_APP1_POPUP_ID, &g->apps, prefs.app1);
	set_db_popup(PANEL_APP2_POPUP_ID, &g->apps, prefs.app2);
	set_db_popup(PANEL_DA1_POPUP_ID, &g->das, prefs.da1);
	set_db_popup(PANEL_DA2_POPUP_ID, &g->das, prefs.da2);
    }
}

static ULong get_db_popup(Word list_id, struct db_list *dbs)
{
    ListPtr list;
    Word index;

    if (dbs->count > 0 && dbs->list != NULL) {
	list = GetObjectPtr(list_id);
	index = LstGetSelection(list);
	if (index < dbs->count) {
	    return dbs->list[index].creator;
	}
    }

    return 0;
}

static void set_prefs()
{
    struct findbtn_prefs prefs;
    struct globals *g;
    ULong creator;
    Byte stack_size;

    get_prefs(&prefs);

    set_pref(PANEL_BUTTON1_LIST_ID, &prefs.button1);
    set_pref(PANEL_BUTTON2_LIST_ID, &prefs.button2);
    set_pref(PANEL_BUTTON3_LIST_ID, &prefs.button3);
    set_pref(PANEL_BUTTON4_LIST_ID, &prefs.button4);

    if (FtrGet(CREATOR, FTR_GLOBALS, (DWord *) &g) == noErr
     && g != NULL) {
	creator = get_db_popup(PANEL_APP1_LIST_ID, &g->apps);
	if (creator != 0) { prefs.app1 = creator; }
	creator = get_db_popup(PANEL_APP2_LIST_ID, &g->apps);
	if (creator != 0) { prefs.app2 = creator; }
	creator = get_db_popup(PANEL_DA1_LIST_ID, &g->das);
	if (creator != 0) { prefs.da1 = creator; }
	creator = get_db_popup(PANEL_DA2_LIST_ID, &g->das);
	if (creator != 0) { prefs.da2 = creator; }
    }

    set_pref(PANEL_LOCALIZER_LIST_ID, &prefs.localizer);

    stack_size = BAD_STACK_SIZE;
    set_pref(PANEL_STACK_LIST_ID, &stack_size);
    if (stack_size != BAD_STACK_SIZE) {
	prefs.stack_size = stack_size * STACK_UNIT;
    }

    PrefSetAppPreferencesV10(CREATOR, PREF_VERSION, &prefs, sizeof prefs);
}

static Int compare(VoidPtr p1, VoidPtr p2, Long other)
{
    return StrCaselessCompare(
	    ((SysDBListItemType *) p1)->name, ((SysDBListItemType *) p2)->name);
}

static void gen_db_list(struct db_list *dbs, ULong type, Boolean looup_name)
{
    int i;

    dbs->ptrs = NULL;

    if (SysCreateDataBaseList(
	    type, 0, &dbs->count, (Handle *) &dbs->h, looup_name) == false) {
	dbs->count = 0;
	dbs->h = NULL;
    } else if ((dbs->list = MemHandleLock(dbs->h)) != NULL
	    && (dbs->ptrs = MemPtrNew(sizeof(CharPtr) * dbs->count)) != NULL) {
	if (looup_name == false) {
	    SysInsertionSort(dbs->list,
		    dbs->count, sizeof(SysDBListItemType), compare, NULL);
	}
	for (i = 0; i < dbs->count; ++i) {
	    dbs->ptrs[i] = dbs->list[i].name;
	}
    }
}

static void free_db_list(struct db_list *dbs)
{
    if (dbs->h != NULL) {
	if (dbs->list != NULL) {
	    MemHandleUnlock(dbs->h);
	}
	MemHandleFree(dbs->h);
    }
    if (dbs->ptrs != NULL) {
	MemPtrFree(dbs->ptrs);
    }
}

Boolean MyHandler(EventPtr event)
{
    Boolean handled = false;
    struct globals *g;
    ListPtr list;

    switch (event->eType) {
    case frmOpenEvent:
	set_popups();
	FrmDrawForm(FrmGetActiveForm());
	g = MemPtrNew(sizeof(struct globals));
	if (g != NULL
	 && FtrSet(CREATOR, FTR_GLOBALS, (DWord) g) == noErr) {
	    gen_db_list(&g->apps, APP_DB_TYPE, true);
	    if (g->apps.count > 0 && g->apps.ptrs != NULL) {
		list = GetObjectPtr(PANEL_APP1_LIST_ID);
		LstSetHeight(list, LIST_HEIGHT(g->apps.count));
		LstSetListChoices(list, g->apps.ptrs, g->apps.count);
		list = GetObjectPtr(PANEL_APP2_LIST_ID);
		LstSetHeight(list, LIST_HEIGHT(g->apps.count));
		LstSetListChoices(list, g->apps.ptrs, g->apps.count);
	    }
	    gen_db_list(&g->das, DA_DB_TYPE, false);
	    if (g->das.count > 0 && g->das.ptrs != NULL) {
		list = GetObjectPtr(PANEL_DA1_LIST_ID);
		LstSetHeight(list, LIST_HEIGHT(g->das.count));
		LstSetListChoices(list, g->das.ptrs, g->das.count);
		list = GetObjectPtr(PANEL_DA2_LIST_ID);
		LstSetHeight(list, LIST_HEIGHT(g->das.count));
		LstSetListChoices(list, g->das.ptrs, g->das.count);
	    }
	}
	set_db_popups();
	handled = true;
	break;

    case frmCloseEvent:
	if (FtrGet(CREATOR, FTR_GLOBALS, (DWord *) &g) == noErr
	 && g != NULL) {
	    free_db_list(&g->apps);
	    free_db_list(&g->das);
	    MemPtrFree(g);
	    FtrUnregister(CREATOR, FTR_GLOBALS);
	}
	break;

    case ctlSelectEvent:
	switch (event->data.ctlSelect.controlID) {
	case PANEL_OK_ID:
	    set_prefs();

	    /* fall through */

	case PANEL_CANCEL_ID:
	    FrmGotoForm(HACKMASTER_FORM_ID);
	    handled = true;
	    break;

	case PANEL_REVERT_ID:
	    set_popups();
	    set_db_popups();
	    handled = true;
	    break;
	}
	break;

    default:		/* to shut up gcc -W ... */
	break;
    }

    return handled;
}
